rm(list=ls())
# this is the initial random seed
set.seed(1234)
# this program serves to estimate genomic selection
# SNP effects and then use these estimates to predict the breeding value of 
# another set of idividuals
# we see the predictive ability by the correlation between estimated breeding value and 
# observed phenotype
# mice data set, body length
# prepared for a course in UBA, Buenos Aires, February 2009
# Andres Legarra, INRA 

solve_densem_pcg_X <- function(X,rhs,vara,vare){
   # solve X'X +I vara sol = rhs by preconditioned conjugate gradient for densem A
   # all effects are random but the mean

   n=length(rhs)
   m=rep(0,n)
   De=rep(1/vara,n)
   De[1]=0
   for (i in 1:n){
     m[i]=1/( crossprod(X[,i])/vare + De[i])
   }
   sol=rep(0,n)
   p=rep(0,n)
   z=rep(0,n)
   w=rep(0,n)
   r=rhs
  
   for (k in 1:1000){
     z=m*r
     tau=crossprod(z,r)
      if (k == 1){
         beta=0
	       p=z
       }else{
         beta=as.numeric(tau/oldtau)
         p=z+beta*p
      }
      w=crossprod(X,X%*%p)/vare + De*p
      alpha=as.numeric(tau/crossprod(p,w))
      sol=sol+alpha*p
      if ((k%%100) != 0){
           r=r-alpha*w
         }else{
           r=rhs-crossprod(X,X%*%sol)/vare - De*sol
      }
      conv=crossprod(r,r)/crossprod(rhs,rhs)
      #print(c('round ',k,'   convergence=',conv))
      if (conv < 1e-14) break
      oldtau=tau
   }   
   #print(c(k,' iterations,   convergence criterion=',conv))
   sol
}

solve_densem_pcg <- function(A,rhs){
   # solve A sol = rhs by preconditioned conjugate gradient for densem A

   n=length(rhs)
   m=rep(0,n)
   m=1/diag(A)
   sol=rep(0,n)
   p=rep(0,n)
   z=rep(0,n)
   w=rep(0,n)
   r=rhs
  
   for (k in 1:1000){
     z=m*r
     tau=crossprod(z,r)
      if (k == 1){
         beta=0
	       p=z
       }else{
         beta=as.numeric(tau/oldtau)
         p=z+beta*p
      }
      w=A%*%p
      alpha=as.numeric(tau/crossprod(p,w))
      sol=sol+alpha*p
      if ((k%%100) != 0){
           r=r-alpha*w
         }else{
           r=rhs-A%*%sol
      }
      conv=crossprod(r,r)/crossprod(rhs,rhs)
      #print(c('round ',k,'   convergence=',conv))
      if (conv < 1e-14) break
      oldtau=tau
   }   
   #print(c(k,' iterations,   convergence criterion=',conv))
   sol
}


  

# this is the data set with all SNPs
# a=matrix( scan('bl10946.txt'),nrow=1829,byrow=TRUE)
# this one has 1 every 10 SNPs
 a=matrix( scan('bl1095.txt'),nrow=1829,byrow=TRUE)
# this one has 1 every 10 SNPs
# a=matrix( scan('bl5473.txt'),nrow=1829,byrow=TRUE)
# extract design matrix for SNPs
X=a[,-(1:3)]
#extract data
y=a[,2]
#center design matrix
for (i in 1:ncol(X)){
  X[,i]=X[,i]-mean(X[,i])
}
# compute trace
trac=sum(diag(tcrossprod(X)))

G = tcrossprod(X)/(trac/1829)
# sustract mean
y=y-mean(y)

vare=0.200
varg=0.040
Z=diag(1829)
ZZ=crossprod(Z)

# set up matrix
LHS=(G*varg)%*%(ZZ/vare)%*%(G*varg)+G*varg
RHS=(G*varg)%*%(t(Z)/vare)%*%y

sol=solve_densem_pcg(LHS,RHS)
# we obtain overall genetic values
a=(G*varg)%*%sol

print(cor(y,a))

