rm(list=ls())
# this is the initial random seed
set.seed(1234)
# this program serves to estimate genomic selection
# SNP effects and then use these estimates to predict the breeding value of 
# another set of idividuals
# we see the predictive ability by the correlation between estimated breeding value and 
# observed phenotype
# mice data set, body length
# prepared for a course in UBA, Buenos Aires, February 2009
# Andres Legarra, INRA 

solve_densem_pcg_X <- function(X,rhs,vara,vare){
   # solve X'X +I vara sol = rhs by preconditioned conjugate gradient for densem A
   # all effects are random but the mean

   n=length(rhs)
   m=rep(0,n)
   De=rep(1/vara,n)
   De[1]=0
   for (i in 1:n){
     m[i]=1/( crossprod(X[,i])/vare + De[i])
   }
   sol=rep(0,n)
   p=rep(0,n)
   z=rep(0,n)
   w=rep(0,n)
   r=rhs
  
   for (k in 1:1000){
     z=m*r
     tau=crossprod(z,r)
      if (k == 1){
         beta=0
	       p=z
       }else{
         beta=as.numeric(tau/oldtau)
         p=z+beta*p
      }
      w=crossprod(X,X%*%p)/vare + De*p
      alpha=as.numeric(tau/crossprod(p,w))
      sol=sol+alpha*p
      if ((k%%100) != 0){
           r=r-alpha*w
         }else{
           r=rhs-crossprod(X,X%*%sol)/vare - De*sol
      }
      conv=crossprod(r,r)/crossprod(rhs,rhs)
      #print(c('round ',k,'   convergence=',conv))
      if (conv < 1e-14) break
      oldtau=tau
   }   
   #print(c(k,' iterations,   convergence criterion=',conv))
   sol
}


GSRU <- function(y,X,vare,vara,ahat=rep(0,dim(X)[2])){
  # this function solves a blup for genomic selection
  # given design matrix X, data y, residual variance vare, SNP variance vara
  # solve by Gauss Seidel
  neq=dim(X)[2]
  e=y-X%*%ahat
  #ahat=rep(0,neq)
  xpx=rep(0,neq)
  epsit=rep(0,1)
  # this is GSRU, legarra & Misztal, Journal of Dairy Science 2008
  #set up diag(X'X)
  for (i in 1:neq) {
    xpx[i]=crossprod(X[,i],X[,i])
  }
  mu=0
  # do until convergence
  for (iter in 1:1000) {
    #Gauss Seidel
    eps=0
    for (i in 1:neq){
      lhs=xpx[i]/vare
      if(i>1) lhs=lhs+1/vara
      rhs=crossprod(X[,i],e)/vare + xpx[i]/vare*ahat[i]
      val=rhs/lhs
      eps=eps+((val - ahat[i])**2)
      e = e - X[,i]*(val - ahat[i])
      ahat[i]=val
    }
    if(iter%%10==0) print(c(iter,eps))
    epsit[iter]=eps
    if(eps<1E-10) break
  }
  print(c(iter,eps,date()))
  ahat
}  

# this is the data set with all SNPs
# a=matrix( scan('bl10946.txt'),nrow=1829,byrow=TRUE)
# this one has 1 every 10 SNPs
 a=matrix( scan('bl1095.txt'),nrow=1829,byrow=TRUE)
# this one has 1 every 10 SNPs
# a=matrix( scan('bl5473.txt'),nrow=1829,byrow=TRUE)
# extract design matrix for SNPs
X=a[,-(1:3)]
#center design matrix
for (i in 1:ncol(X)){
  X[,i]=X[,i]-mean(X[,i])
}
# compute trace
trac=sum(diag(tcrossprod(X)))

#extract data
y=a[,2]
# add design matrix for general mean
X=cbind(1,X)
vare=0.200
#varg=0.040
#vara=9.09E-06 # this is for the whole 10946 snps
vara=1.02E-04 # this is for 1095 snps
#vara=4.5E-04 # this is for 5473 SNPs
#vara=varg/(trac/1829)
print(c("vara",vara))
a = GSRU(y,X,vare,vara)
#a = solve_densem_pcg_X(X,crossprod(X,y)/vare,vara,vare)

yhat=X%*%a
print(cor(y,yhat))

# try this with ind_in_training = 200, 1000, 1600
# and for the case of 1600 with vara= 1E-06, 1E-04, 1E-02

# now do the same by splitting the data into two
ind_in_training=200
vare=0.200
vara=1.02E-04

# here we select some animals to be the "training" data set
training=sample(1:1829,ind_in_training)
# tho others will be used for validation
valid=(1:1829)[-training]
# we estimate SNP effects with individuals in "training" 
# excluding the ones in validation
a = GSRU(y[training],X[training,],vare,vara,a)
#a=solve_densem_pcg_X(X[training,],crossprod(X[training,],y[training])/vare,vara,vare)
# using estimates for a, here we predict phenotypes for everybody
yhat=X%*%a
# how much is the accuracy for animals in validation?
print(cor(y[valid],yhat[valid]))
# how much is the accuracy for animals in training?
print(cor(y[training],yhat[training]))
   

